#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from copy import deepcopy
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import re
import numpy as np
import yaml
from kgforge.core.commons.files import load_file_as_byte
from pandas import DataFrame
from rdflib import Graph
from urllib.parse import quote_plus, urlparse

from kgforge.core import Resource
from kgforge.core.archetypes import Mapping, Model, Resolver, Store
from kgforge.core.commons.context import Context
from kgforge.core.commons.actions import LazyAction
from kgforge.core.commons.dictionaries import with_defaults
from kgforge.core.commons.exceptions import ResolvingError
from kgforge.core.commons.execution import catch
from kgforge.core.commons.imports import import_class
from kgforge.core.commons.strategies import ResolvingStrategy
from kgforge.core.commons.formatter import Formatter
from kgforge.core.conversions.dataframe import as_dataframe, from_dataframe
from kgforge.core.conversions.json import as_json, from_json
from kgforge.core.conversions.rdf import (
    as_jsonld,
    from_jsonld,
    as_graph,
    from_graph,
    Form,
)
from kgforge.core.reshaping import Reshaper
from kgforge.core.wrappings.paths import PathsWrapper, wrap_paths
from kgforge.specializations.mappers import DictionaryMapper
from kgforge.specializations.mappings import DictionaryMapping


class KnowledgeGraphForge:

    # POLICY Class name should be imported in the corresponding module __init__.py.

    # No catching of exceptions so that no incomplete instance is created if an error occurs.
    # This is a best practice in Python for __init__().
    def __init__(self, configuration: Union[str, Dict], **kwargs) -> None:
        """
        Configure and create a Knowledge Graph forge session.
        See https://github.com/BlueBrain/nexus-forge/blob/master/examples/notebooks/use-cases/prod-forge-nexus.yml for an example of configuration.

        Required minimal configuration: name, origin and source for Model and name for Store.
        Keyword arguments could be used to override the configuration provided for the Store.

        The configuration could be provided either as:

        - inline YAML or a path to a YAML file with the following structure:

         Model:
           name: <a class name of a Model>
           origin: <'directory', 'url', or 'store'>
           source: <a directory path, an URL, or the class name of a Store>
           bucket: <when 'origin' is 'store', a Store bucket>
           endpoint: <when 'origin' is 'store', a Store endpoint, default to Store:endpoint>
           token: <when 'origin' is 'store', a Store token, default to Store:token>
           context:
             iri: <an IRI>
             bucket: <when 'origin' is 'store', a Store bucket, default to Model:bucket>
             endpoint: <when 'origin' is 'store', a Store endpoint, default to Model:endpoint>
             token: <when 'origin' is 'store', a Store token, default to Model:token>

         Store:
           name: <a class name of a Store>
           endpoint: <an URL>
           bucket: <a bucket as a string>
           token: <a token as a string>
           searchendpoints:
             <querytype>: <a query paradigm supported by configured store (e.g. sparql)>
               endpoint: <an IRI of a query endpoint>
           params:
               <Store method>: <e.g. register, tag, ...>
                   param: <http query param value to use for the Store method>
           versioned_id_template: <a string template using 'x' to access resource fields>
           file_resource_mapping: <an Hjson string, a file path, or an URL>

         Resolvers:
           <scope>:
             - resolver: <a class name of a Resolver>
               origin: <'directory', 'web_service', or 'store'>
               source: <a directory path, a web service endpoint, or the class name of a Store>
               targets:
                 - identifier: <a name, or an IRI>
                   bucket: <a file name, an URL path, or a Store bucket>
                   filters:
                     - path: <a resource property path>
                     - value: <a resource property value to filter with>
               searchendpoints:
                 sparql:
                   endpoint: <A SPARQL endpoint to send resolving query to. Only used for resolvers based on SPARQL>
               resolve_with_properties: <a list of str currently only supported by DemoResolver>
               result_resource_mapping: <an Hjson string, a file path, or an URL>
               endpoint: <when 'origin' is 'store', a Store endpoint, default to Store:endpoint>
               token: <when 'origin' is 'store', a Store token, default to Store:token>

         Formatters:
           <identifier>: <a string template with replacement fields delimited by braces, i.e. '{}'>

        - A Python dictionary with the following structure:

         {
             "Model": {
                 "name": <str>,
                 "origin": <str>,
                 "source": <str>,
                 "bucket": <str>,
                 "endpoint": <str>,
                 "token": <str>,
                 "context": {
                       "iri": <str>,
                       "bucket": <str>,
                       "endpoint": <str>,
                       "token": <str>,
                 }
             },
             "Store": {
                 "name": <str>,
                 "endpoint": <str>,
                 "bucket": <str>,
                 "token": <str>,
                 "searchendpoints": {
                   "<querytype>": {
                       "endpoint": <str>
                   }
                 },
                 "params": {
                   "<Store method>": {
                       "param": <str>
                   }
                 },
                 "versioned_id_template": <str>,
                 "file_resource_mapping": <str>,
             },
             "Resolvers": {
                 "<scope>": [
                     {
                         "resolver": <str>,
                         "origin": <str>,
                         "source": <str>,
                         "targets": [
                             {
                                 "identifier": <str>,
                                 "bucket": <str>,
                                 "filter":[
                                    {
                                        "path": <str>,
                                        "value": <str>
                                    }
                                 ]
                             },
                             ...,
                         ],
                         "searchendpoints":{
                            "sparql":{
                                "endpoint": <str>
                            }
                         },
                         "resolve_with_properties": [str]
                         "result_resource_mapping": <str>,
                         "endpoint": <str>,
                         "token": <str>,
                     },
                     ...,
                 ],
             },
             "Formatters": {
                 "<name>": <str>,
                 ...,
             },
         }

         In the configuration, Class name could be provided in three formats:
               * 'SomeClass',
               * 'SomeClass from package.module',
               * 'SomeClass from package.module.file'.
         When the class is from this package, the first is used. Otherwise, the two others.
         Values using braces with something inside should be quoted with double quotes.

        :param configuration: a configuration content or file path
        :param kwargs:  keyword arguments
        """

        if isinstance(configuration, str):
            config_data = load_file_as_byte(configuration)
            config_data = config_data.decode("utf-8")
            config = yaml.safe_load(config_data)
        else:
            config = deepcopy(configuration)

        # Debugging.
        self._debug = kwargs.pop("debug", False)

        # Store.
        store_config = config.pop("Store")
        store_config.update(kwargs)

        # Model.
        model_config = config.pop("Model")
        if model_config["origin"] == "store":
            with_defaults(
                model_config,
                store_config,
                "source",
                "name",
                ["endpoint", "token", "bucket", "vocabulary"],
            )
        model_name = model_config.pop("name")
        model = import_class(model_name, "models")
        self._model: Model = model(**model_config)

        # Store.
        store_config.update(model_context=self._model.context())
        store_name = store_config.pop("name")
        store = import_class(store_name, "stores")
        self._store: Store = store(**store_config)
        store_config.update(name=store_name)

        # Resolvers.
        resolvers_config = config.pop("Resolvers", None)
        # Format: Optional[Dict[scope_name, Dict[resolver_name, Resolver]]].
        self._resolvers: Optional[Dict[str, Dict[str, Resolver]]] = (
            prepare_resolvers(resolvers_config, store_config)
            if resolvers_config
            else None
        )

        # Formatters.
        self._formatters: Optional[Dict[str, str]] = config.pop("Formatters", None)

    @catch
    def prefixes(self, pretty: bool = True) -> Optional[Dict[str, str]]:
        """
        Print (pretty=True) prefix mappings:
            Used prefixes:
            rdf          http://www.w3.org/1999/02/22-rdf-syntax-ns#
            prov         http://www.w3.org/ns/prov#

        Or return (pretty=False) them as dictionary:
            {
                "rdf":"http://www.w3.org/1999/02/22-rdf-syntax-ns#",
                "prov":"http://www.w3.org/ns/prov#"
            }

        :param pretty: a boolean
        :return: Optional[Dict[str, str]]
        """
        return self._model.prefixes(pretty)

    @catch
    def types(self, pretty: bool = True) -> Optional[List[str]]:
        """
        Print (pretty=True) known resource types (by the configured Model):

        Managed entity types:
            - Dataset
            - Entity

        Or return (pretty=False) them as a list of strings:

        ["Dataset", "Entity"]

        :param pretty: a boolean
        :return: Optional[List[str]]
        """
        return self._model.types(pretty)

    @catch
    def template(
        self, type: str, only_required: bool = False, output: str = "hjson"
    ) -> Optional[Dict]:
        """
        Print the schema associated with a given resource type (must be listed in forge.types(...)) in hjson (output='hjson') or JSON (output='json') format.
        When output='dict', a dictionary is returned.

        :param type: a resource type
        :param only_required: whether to retrieve required properties (True) or all properties (False) for a resource type
        :param output: whether to print ('hjson' or 'json') the template or return it as dictionary ('dict')
        :return: Optional[Dict]
        """
        return self._model.template(type, only_required, output)

    # No @catch because the error handling is done by execution.run().
    def validate(
        self,
        data: Union[Resource, List[Resource]],
        execute_actions_before: bool=False,
        type_: str=None
    ) -> None:
        """
        Check if resources conform to their corresponding schemas. This method will try to infer the schema of a resource from its type.
        For this method to work, a provided resource should have a type property which is listed in forge.type(...).
        It is not possible to validate a resource with a LazyAction as value of one of its property.
        The LazyAction has to be executed first (execute_actions_before=True) before validation. A report is printed after the validation is performed.

        :param data: a resource or a list of resources to validate
        :param execute_actions_before: whether to execute a LazyAction value of one of a resource property (True) or not (False) prior to validation
        :param type_: the type to validate the data against it. If None, the validation function will look for a type attribute in the Resource
        :return: None
        """
        self._model.validate(data, execute_actions_before, type_=type_)

    # Resolving User Interface.

    def resolvers(self, output: str = "print") -> Optional[Dict]:
        """
        Print (output='print') the available resolvers or return them based as dictionary (output='dict').

        The dictionary returned has the following format:
        {
            "firstResolver": {
                "firstTarget": {
                    "bucket": "firstSource",
                    "filters":[
                        {
                            "path": "path",
                            "value": "value"
                        }
                    ]
                },
                "secondTarget": {
                    "bucket": "secondSource",
                }
            }
        }

        :param output: defines whether the resolvers should printed or returned as a dictionary
        :return: Optional[Dict]
        """
        if output == "print":
            print("Available scopes:")
            for scope, scope_value in sorted(self._resolvers.items()):
                print(" - ", scope, ":")
                for resolver_key, resolver_value in scope_value.items():
                    print("     - resolver: ", resolver_key)
                    print(
                        "         - targets: ", ",".join(resolver_value.targets.keys())
                    )
        elif output == "dict":
            resolvers_dict = dict()
            # iterate over target_key, target_value and fill dictionary with targets
            for scope, scope_value in sorted(self._resolvers.items()):
                individual_dict = dict()
                for resolver_key, resolver_value in scope_value.items():
                    for target_key, target_value in resolver_value.targets.items():
                        individual_dict[target_key] = target_value
                resolvers_dict[scope] = individual_dict
            return resolvers_dict
        else:
            raise ValueError("unrecognized output")

    @catch
    def resolve(
        self,
        text: Union[str, List[str], Resource],
        scope: Optional[str] = None,
        resolver: Optional[str] = None,
        target: Optional[str] = None,
        type: Optional[str] = None,
        strategy: Union[ResolvingStrategy, str] = ResolvingStrategy.BEST_MATCH,
        resolving_context: Optional[Any] = None,
        property_to_resolve: Optional[str] = None,
        merge_inplace_as: Optional[str] = None,
        limit: Optional[int] = 10,
        threshold: Optional[float] = 0.5,
    ) -> Optional[Union[Resource, List[Resource], Dict[str, List[Resource]]]]:
        """
        Resolve text(s) or a resource into existing resources (from the configured Store) depending on the resolving strategy.
        Returned resources are sorted (best first) by score. See more details in the docs: https://nexus-forge.readthedocs.io/en/latest/interaction.html#resolving.

        :param text: text(s) or resources to resolve
        :param scope: a scope identifier
        :param resolver: a resolver name
        :param target: a target identifier
        :param type: a type of resources to resolve
        :param strategy:  the ResolvingStrategy to apply: BEST_MATCH (return the resource with the best score), EXACT_MATCH (return the resource with the highest score), ALL_MATCHES (return all retrieved resources sorted by score)
        :param resolving_context: A context (e.g JSON-LD context) to use during resolving
        :param property_to_resolve: the property from which the text(s) to resolve is taken when 'text' is a Resource
        :param merge_inplace_as: the property name that should hold the resolving result when 'text' is a Resource. When missing, the resolving result is returned insteada of being added to the provided resource.
        :param limit: the number of resources to retrieve
        :param threshold: the threshold score
        :return: Optional[Union[Resource, List[Resource], Dict[str, List[Resource]]]]
        """
        if self._resolvers is not None:
            if scope is not None and scope in self._resolvers:
                if resolver is not None:
                    rov = self._resolvers[scope][resolver]
                else:
                    scope_resolvers = list(self._resolvers[scope].values())
                    if len(scope_resolvers) == 1:
                        rov = scope_resolvers[0]
                    else:
                        raise ResolvingError("resolver name should be specified")
            elif scope is not None:
                raise AttributeError(
                    f"{scope} is not a configured resolving scope. Configured scopes are {list(self._resolvers.keys())}"
                )
            else:
                scopes_resolvers = list(self._resolvers.values())
                if resolver is not None:
                    if len(self._resolvers) == 1:
                        rov = scopes_resolvers[0][resolver]
                    else:
                        raise ResolvingError(
                            "resolving scope should be specified in case of multiple available resolvers"
                        )
                else:
                    if len(self._resolvers) == 1 and len(scopes_resolvers[0]) == 1:
                        rov = list(scopes_resolvers[0].values())[0]
                    else:
                        raise ResolvingError(
                            "resolving scope or resolver name should be "
                            "specified in case of multiple available resolvers"
                        )
            try:
                strategy = (
                    strategy
                    if isinstance(strategy, ResolvingStrategy)
                    else ResolvingStrategy[strategy]
                )
            except Exception as e:
                raise AttributeError(
                    f"Invalid ResolvingStrategy value '{strategy}'. Allowed names are {[name for name, member in ResolvingStrategy.__members__.items()]} and allowed members are {[member for name, member in ResolvingStrategy.__members__.items()]}"
                )
            return rov.resolve(
                text,
                target,
                type,
                strategy,
                resolving_context,
                property_to_resolve,
                merge_inplace_as,
                limit,
                threshold,
            )
        else:
            raise ResolvingError("no resolvers have been configured")

    # Formatting User Interface.

    @catch
    def format(self, what: str = None, *args, formatter: Union[Formatter, str] = Formatter.STR, uri: str = None, **kwargs) -> str:
        """
        Select a configured formatter (see https://nexus-forge.readthedocs.io/en/latest/interaction.html#formatting) string (identified by 'what') and format it using provided '*args'
        :param what: a configured str format name. Required formatter:str = Formatter.STR
        :param args: a list of string to use for formatting
        :param formatter: the type of formatter to use. STR (default, corresponds to str.format(*args, **kwargs)), URI_REWRITER (Store based URI rewritter)  
        :param uri: a URI to rewrite. Required formatter:str = Formatter.URI_REWRITER
        :return: str
        """
        
        if what and uri:
            raise AttributeError(
                    f"both 'what': {what} and 'uri': {uri} arguments are provided. One of them should be provided."
            )

        try:
            formatter = (
                formatter
                if isinstance(formatter, Formatter)
                else Formatter[formatter]
            )
        except Exception as e:
            raise AttributeError(
                f"Invalid Formatter value '{formatter}'. Allowed names are {[name for name, member in Formatter.__members__.items()]} and allowed members are {[member for name, member in Formatter.__members__.items()]}"
            )
        
        if formatter == Formatter.STR:
            if what is None:
                raise AttributeError(
                        f"A non None 'what' value is required when formatter == Formatter.STR"
                )
            return self._formatters[what].format(*args, **kwargs)
        elif formatter == Formatter.URI_REWRITER:
            if uri is None:
                raise AttributeError(
                            f"A non None 'uri' value is required when formatter == Formatter.URI_REWRITER"
                    )
            return self._store.rewrite_uri(uri, self.get_store_context(), **kwargs)
        else:
            raise AttributeError(
                    f"{formatter} is not a valid formatter. Valid formatters are {[fm.value for fm in Formatter]}"
            )


    # Mapping User Interface.

    @catch
    def sources(self, pretty: bool = True) -> Optional[List[str]]:
        """
        Print(pretty=True) or return (pretty=False) configured data sources.
        :param pretty: a boolean
        :return: Optional[List[str]]
        """
        return self._model.sources(pretty)

    @catch
    def mappings(
        self, source: str, pretty: bool = True
    ) -> Optional[Dict[str, List[str]]]:
        """
        Print(pretty=True) or return (pretty=False) configured mappings for a given source.
        Mappings are declarative rules encoding the logic of transforming data from a given source and format into resources.

        :param source: the source for which mappings are return
        :param pretty: a boolean
        :return: Optional[Dict[str, List[str]]]
        """
        return self._model.mappings(source, pretty)

    @catch
    def mapping(
        self, entity: str, source: str, type: Callable = DictionaryMapping
    ) -> Mapping:
        """
        Return a Mapping object of type 'type' for a resource type 'entity' and a source.

        :param entity: a resource type
        :param source: a data source
        :param type: a Mapping class
        :return: Mapping
        """
        return self._model.mapping(entity, source, type)

    @catch
    def map(
        self,
        data: Any,
        mapping: Union[Mapping, List[Mapping]],
        mapper: Callable = DictionaryMapper,
        na: Union[Any, List[Any]] = None,
    ) -> Union[Resource, List[Resource]]:
        """
        Transform data to resources using transformations rules provided as mappings. The format of the data to transform
        should be supported by the provided Mapper.

        :param data: data to be mapped
        :param mapping: mappings to use
        :param mapper: Mapper class to use
        :param na: represents missing values
        :return: Union[Resource, List[Resource]]
        """
        return mapper(self).map(data, mapping, na)

    # Reshaping User Interface.

    @catch
    def reshape(
        self,
        data: Union[Resource, List[Resource]],
        keep: List[str],
        versioned: bool = False,
    ) -> Union[Resource, List[Resource]]:
        """
        Keep only a provided list of properties ('keep') from a resource of list of resources.

        :param data: the resources to reshape
        :param keep: the properties to keep
        :param versioned: whether to use versioned identifier (True) or not (False)
        :return: Union[Resource, List[Resource]]
        """
        return Reshaper(self._store.versioned_id_template).reshape(
            data, keep, versioned
        )

    # Querying User Interface.

    @catch
    def retrieve(
        self,
        id: str,
        version: Optional[Union[int, str]] = None,
        cross_bucket: bool = False,
        **params
    ) -> Resource:
        """
        Retrieve a resource by its identifier from the configured store and possibly at a given version.

        :param id: the resource identifier to retrieve
        :param version: a version of the resource to retrieve
        :param cross_bucket: instructs the configured store to whether search beyond the configured bucket (True) or not (False)
        :param params: a dictionary of parameters.
        :return: Resource
        """
        return self._store.retrieve(id, version, cross_bucket, **params)

    @catch
    def paths(self, type: str) -> PathsWrapper:
        """
        Return a PathsWrapper for a given resource type. A PathsWrapper object wraps a resource properties tree as a set of navigable paths.

        :param type: a resource type
        :return: PathsWrapper
        """
        template = self._model.template(type, False, "dict")
        return wrap_paths(template)

    @catch
    def search(self, *filters, **params) -> List[Resource]:
        """
        Search for resources based on a list of filters. The search results can be controlled (e.g. number of results) by setting parameters.
        See docs for more details: https://nexus-forge.readthedocs.io/en/latest/interaction.html#querying

        :param filters: a list of filters
        :param params: a dictionary of parameters
        :return: List[Resource]
        """
        resolvers = (
            list(self._resolvers.values()) if self._resolvers is not None else None
        )
        return self._store.search(resolvers, *filters, **params)

    @catch
    def sparql(
        self,
        query: str,
        debug: bool = False,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        **params
    ) -> List[Resource]:
        """
        Search for resources using a SPARQL query. See SPARQL docs: https://www.w3.org/TR/sparql11-query.

        :param query: a SPARQL query
        :param debug: a boolean
        :param limit: the number of resources to retrieve. Default to 100. If provided in query limit will be replaced
        :param offset: how many results to skip from the first one. If provided in query offset will be replaced
        :param params: a dictionary of parameters. Supported params are: rewrite (whether to rewrite the sparql query or run it as is)
        :return: List[Resource]
        """
        return self._store.sparql(query, debug, limit, offset, **params)

    @catch
    def elastic(
        self,
        query: str,
        debug: bool = False,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> List[Resource]:
        """
        Search for resources using an ElasticSearch DSL query. See ElasticSearch DSL docs: https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl.html.

        :param query: an ElasticSerach DSL query
        :param debug: a boolean
        :param limit: the number of resources to retrieve
        :param offset: how many results to skip from the first one
        :return: List[Resource]
        """
        return self._store.elastic(query, debug, limit, offset)

    @catch
    def download(
        self,
        data: Union[Resource, List[Resource]],
        follow: str = "distribution.contentUrl",
        path: str = ".",
        overwrite: bool = False,
        cross_bucket: bool = False,
        content_type: str = None
    ) -> None:
        """
        Download files attached to a resource or a list of resources.

        :param data: the resources whose attached files to download
        :param follow: the property path holding a URL to download the files
        :param path: where to output the downloaded files
        :param overwrite: whether to replace (True) and existing file with the same name or not (False)
        :param cross_bucket: instructs the configured store to whether download files beyond the configured bucket (True) or not (False)
        :param content_type: the content_type of the files to download
        """
        # path: DirPath.
        self._store.download(data, follow, path, overwrite, cross_bucket, content_type)

    # Storing User Interface.

    # No @catch because the error handling is done by execution.run().
    def register(
        self, data: Union[Resource, List[Resource]], schema_id: Optional[str] = None
    ) -> None:
        """
        Store a resource or list of resources in the configured Store.

        :param data: the resources to register
        :param schema_id: an identifier of the schema the registered resources should conform to
        """
        #self._store.mapper = self._store.mapper(self)
        self._store.register(data, schema_id)

    # No @catch because the error handling is done by execution.run().
    def update(
        self, data: Union[Resource, List[Resource]], schema_id: Optional[str] = None
    ) -> None:
        """
        Update a resource or a list of resources in the configured Store.

        :param data: the resources to update
        :param schema_id: an identifier of the schema the updated resources should conform to
        """
        self._store.update(data, schema_id)

    # No @catch because the error handling is done by execution.run().
    def deprecate(self, data: Union[Resource, List[Resource]]) -> None:
        """
        Deprecate a resource or a list of resources.

        :param: the resources to deprecate
        """
        self._store.deprecate(data)

    # Versioning User Interface.

    # No @catch because the error handling is done by execution.run().
    def tag(self, data: Union[Resource, List[Resource]], value: str) -> None:
        """
        Assign a tag (value) to a resource or a list of resources. A tag can be seen as a version.

        :param data: the resources to tag
        :param value: the tag value
        """
        self._store.tag(data, value)

    # No @catch because the error handling is done by execution.run().
    def freeze(self, data: Union[Resource, List[Resource]]) -> None:
        """
        Replace all resources' references within the provided resources with a versioned identifier.
        See Versioning docs: https://nexus-forge.readthedocs.io/en/latest/interaction.html#versioning.

        :param data: the resources to freeze
        """
        self._store.freeze(data)

    # Files Handling User Interface.

    @catch
    def attach(self, path: str, content_type: str = None) -> LazyAction:
        """
        Return a LazyAction for future upload  of files located in a provided path.
        The output of this method can be used to attach files to a resource: https://nexus-forge.readthedocs.io/en/latest/interaction.html#resource.

        :param path: path to upload files from
        :param content_type: the content_type of the files to upload
        :return: LazyAction
        """
        # path: Union[FilePath, DirPath].
        return LazyAction(self._store.upload, path, content_type, self)

    # Converting User Interface.

    @catch
    def as_json(
        self,
        data: Union[Resource, List[Resource]],
        expanded: bool = False,
        store_metadata: bool = False,
    ) -> Union[Dict, List[Dict]]:
        """
        Convert a resource or a list of resources to JSON.

        :param data: the resources to convert
        :param expanded: whether to expand (True) the JSON keys as URIs using a JSON-LD context (if any) or not (False)
        :param store_metadata: whether to add (True) store related metadata (e.g rev) to the output or not (False)
        :return: Union[Dict, List[Dict]]
        """
        return as_json(
            data,
            expanded,
            store_metadata,
            self._model.context(),
            self._store.metadata_context,
            self._model.resolve_context,
        )

    @catch
    def as_jsonld(
        self,
        data: Union[Resource, List[Resource]],
        form: str = Form.COMPACTED.value,
        store_metadata: bool = False,
        **params
    ) -> Union[Dict, List[Dict]]:
        """
        Convert a resource or a list of resources to JSON-LD.

        :param data: the resources to convert
        :param form: whether to expand ('expanded') or compact ('compacted') the JSON-LD output
        :param store_metadata: whether to add (True) store related metadata (e.g rev) to the output or not (False)
        :param params: a dictionary of parameters.
        :return: Union[Dict, List[Dict]]
        """
        return as_jsonld(
            data,
            form,
            store_metadata,
            self._model.context(),
            self._store.metadata_context,
            self._model.resolve_context,
            **params
        )

    @catch
    def as_graph(
        self, data: Union[Resource, List[Resource]], store_metadata: bool = False
    ) -> Graph:
        """
        Convert a resource or a list of resources to a RDFLib Graph object: https://rdflib.readthedocs.io/en/stable/intro_to_graphs.html.

        :param data: the resources to convert
        :param store_metadata: whether to add (True) store related metadata (e.g rev) to the output or not (False)
        :return: rdflib.Graph
        """
        return as_graph(
            data,
            store_metadata,
            self._model.context(),
            self._store.metadata_context,
            self._model.resolve_context,
        )

    @catch
    def as_dataframe(
        self,
        data: Union[Resource, List[Resource]],
        na: Union[Any, List[Any]] = [None],
        nesting: str = ".",
        expanded: bool = False,
        store_metadata: bool = False,
    ) -> DataFrame:
        """
        Convert a resource or a list of resources to pandas.DataFrame.

        :param data: the resources to convert
        :param na: represents missing values
        :param nesting: str to use to join when flattening nested properties as columns
        :param expanded: whether to expand (True) resources' properties as URIs using a JSON-LD context (if any) or not (False)
        :param store_metadata: whether to add (True) store related metadata (e.g rev) to the output or not (False)
        :return: pandas.DataFrame
        """
        return as_dataframe(
            data,
            na,
            nesting,
            expanded,
            store_metadata,
            self._model.context(),
            self._store.metadata_context,
            self._model.resolve_context,
        )

    @catch
    def from_json(
        self, data: Union[Dict, List[Dict]], na: Union[Any, List[Any]] = None
    ) -> Union[Resource, List[Resource]]:
        """
        Convert a JSON document or a list of JSON documents to a resource or a list of resources.

        :param data: the JSON documents to convert
        :param na: represents missing values
        :return: Union[Resource, List[Resource]]
        """
        return from_json(data, na)

    @catch
    def from_jsonld(
        self, data: Union[Dict, List[Dict]]
    ) -> Union[Resource, List[Resource]]:
        """
        Convert a JSON-LD document or a list of JSON-LD documents to a resource or a list of resources.

        :param data: the JSON-LD documents to convert
        :return: Union[Resource, List[Resource]]
        """
        return from_jsonld(data)

    @catch
    def from_graph(
        self,
        data: Graph,
        type: Union[str, List[str]] = None,
        frame: Dict = None,
        use_model_context=False,
    ) -> Union[Resource, List[Resource]]:
        """
        Convert a RDFLib.Graph object to a resource or a list of resources. What to convert from the RDFLib.Graph can be
        selected by type and how to shape them as resources can be controlled using a JSON-LD Frame: https://www.w3.org/TR/json-ld11-framing.

        :param data: the RDFLib.Graph object to convert
        :param type: the types of RDF resources to convert from the RDFLib.Graph object
        :param frame: the JSON-LD frame to apply during conversion
        :param use_model_context: whether to use  (True) the JSON-LD context from the configured Model or not (False)
        :return: Union[Resource, List[Resource]]
        """
        context = self._model.context() if use_model_context else None
        return from_graph(data, type, frame, context)

    @catch
    def from_dataframe(
        self, data: DataFrame, na: Union[Any, List[Any]] = np.nan, nesting: str = "."
    ) -> Union[Resource, List[Resource]]:
        """
        Convert a pandas.DataFrame to a resource or a list of resources.

        :param data: the pandas.DataFrame to convert
        :param na: represents missing values
        :param nesting: str to use to detect nested nested properties
        :return: Union[Resource, List[Resource]]
        """
        return from_dataframe(data, na, nesting)
    
    def get_store_context(self):
        """Expose the context used in the store."""
        return self._store.context

    def get_model_context(self):
        """Expose the context used in the model."""
        return self._model.context()
    
def prepare_resolvers(
    config: Dict, store_config: Dict
) -> Dict[str, Dict[str, Resolver]]:
    return {
        scope: dict(prepare_resolver(x, store_config) for x in configs)
        for scope, configs in config.items()
    }


def prepare_resolver(config: Dict, store_config: Dict) -> Tuple[str, Resolver]:
    if config["origin"] == "store":
        with_defaults(
            config,
            store_config,
            "source",
            "name",
            [
                "endpoint",
                "token",
                "bucket",
                "model_context",
                "searchendpoints",
                "vocabulary",
            ],
        )
    resolver_name = config.pop("resolver")
    resolver = import_class(resolver_name, "resolvers")
    return resolver.__name__, resolver(**config)



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

# Test suite for initializing a forge.

import pytest

from kgforge.core import KnowledgeGraphForge

SCOPE = "terms"
MODEL = "DemoModel"
STORE = "DemoStore"
RESOLVER = "DemoResolver"


class TestForgeInitialization:

    def test_initialization(self, config):
        forge = KnowledgeGraphForge(config)
        assert type(forge._model).__name__ == MODEL
        assert type(forge._store).__name__ == STORE
        assert type(forge._resolvers[SCOPE][RESOLVER]).__name__ == RESOLVER


class TestResolver:
    """
    Tests the .resolver() function
    """

    def test_resolver_returns_dict(self, config):
        forge = KnowledgeGraphForge(config)
        resolvers_dict = forge.resolvers(output="dict")
        assert type(resolvers_dict) is dict

    def test_resolver_does_not_return_if_print(self, config):
        forge = KnowledgeGraphForge(config)
        resolvers_dict = forge.resolvers(output="print")
        assert resolvers_dict is None

    def test_resolver_returns_correct_dictionary(self, config):
        dict_result = {'terms': {'sex': {'bucket': 'sex.json', 'filters':{'label': 'female', 'type': 'class'}}}}
        forge = KnowledgeGraphForge(config)
        resolvers_dict = forge.resolvers(output="dict")
        assert resolvers_dict == dict_result



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

# Placeholder for the generic parameterizable test suite for resolvers.
from kgforge.core.archetypes.resolver import _build_resolving_query
from kgforge.core.commons.sparql_query_builder import SPARQLQueryBuilder
import pytest

from kgforge.core.commons.strategies import ResolvingStrategy

@pytest.mark.parametrize(
        "filters, query_template, properties_to_filter_with, text, expected_query",
        [
            pytest.param(
                {"path_1":"value_1"},
                ("""
        CONSTRUCT {{
            ?id a ?type ;
            label ?label ;
            prefLabel ?prefLabel ;
            altLabel ?altLabel ;
            definition ?definition;
            subClassOf ?subClassOf ;
            isDefinedBy ?isDefinedBy ;
            notation ?notation
        }} WHERE {{
            ?id a ?type ;
                label ?label ; 
            OPTIONAL {{
            ?id subClassOf ?subClassOf ;
            }}
            OPTIONAL {{
            ?id definition ?definition ;
            }}
            OPTIONAL {{
            ?id prefLabel ?prefLabel .
            }}
            OPTIONAL {{
            ?id altLabel ?altLabel .
            }}
            OPTIONAL {{
            ?id isDefinedBy ?isDefinedBy .
            }}     
            OPTIONAL {{
            ?id notation ?notation .
            }}    
            {{
            SELECT * WHERE {{
                {{ {0} ; label ?label {1} }} UNION
                {{ {0} ; notation ?notation {2} }} UNION
                {{ {0} ; prefLabel ?prefLabel {3} }} UNION
                {{ {0} ; altLabel ?altLabel {4} }}
            }} LIMIT {5}
            }}
        }}
        """),
        (['label', 'notation', 'prefLabel', 'altLabel']),
        ('L6_BC'),
                (
                    
                """
        CONSTRUCT {
            ?id a ?type ;
            label ?label ;
            prefLabel ?prefLabel ;
            altLabel ?altLabel ;
            definition ?definition;
            subClassOf ?subClassOf ;
            isDefinedBy ?isDefinedBy ;
            notation ?notation
        } WHERE {
            ?id a ?type ;
                label ?label ; 
            OPTIONAL {
            ?id subClassOf ?subClassOf ;
            }
            OPTIONAL {
            ?id definition ?definition ;
            }
            OPTIONAL {
            ?id prefLabel ?prefLabel .
            }
            OPTIONAL {
            ?id altLabel ?altLabel .
            }
            OPTIONAL {
            ?id isDefinedBy ?isDefinedBy .
            }     
            OPTIONAL {
            ?id notation ?notation .
            }    
            {
            SELECT * WHERE {
                { ?id <https://bluebrain.github.io/nexus/vocabulary/deprecated> "false"^^xsd:boolean ; a Class ; 
 path_1 ?v0 . 
 FILTER(?v0 = "value_1") ; label ?label  FILTER (?label = "L6_BC") } UNION
                { ?id <https://bluebrain.github.io/nexus/vocabulary/deprecated> "false"^^xsd:boolean ; a Class ; 
 path_1 ?v0 . 
 FILTER(?v0 = "value_1") ; notation ?notation  FILTER (?notation = "L6_BC") } UNION
                { ?id <https://bluebrain.github.io/nexus/vocabulary/deprecated> "false"^^xsd:boolean ; a Class ; 
 path_1 ?v0 . 
 FILTER(?v0 = "value_1") ; prefLabel ?prefLabel  FILTER (?prefLabel = "L6_BC") } UNION
                { ?id <https://bluebrain.github.io/nexus/vocabulary/deprecated> "false"^^xsd:boolean ; a Class ; 
 path_1 ?v0 . 
 FILTER(?v0 = "value_1") ; altLabel ?altLabel  FILTER (?altLabel = "L6_BC") }
            } LIMIT 1
            }
        }
        """
                ),
                id="ontology_query_str"
            ),
            pytest.param(
                {"path_1":"value_1"},
                (
                """
            CONSTRUCT {{
                ?id a ?type ;
                name ?name ;
                givenName ?givenName ;
                familyName ?familyName
            }} WHERE {{
              ?id a ?type . 
              OPTIONAL {{
                ?id name ?name .
              }}
              OPTIONAL {{
                ?id givenName ?givenName . 
              }}
              OPTIONAL {{
                ?id familyName ?familyName .
              }}
              {{
                SELECT * WHERE {{
                  {{ {0} ; name ?name {1} }} UNION
                  {{ {0} ; familyName ?familyName; givenName ?givenName {2} }} UNION
                  {{ {0} ; familyName ?familyName; givenName ?givenName {3} }}
                }} LIMIT {4}
              }}
            }}
            """),
            (['name', 'givenName', 'familyName']),
            ("EPFL"),
                ("""
            CONSTRUCT {
                ?id a ?type ;
                name ?name ;
                givenName ?givenName ;
                familyName ?familyName
            } WHERE {
              ?id a ?type . 
              OPTIONAL {
                ?id name ?name .
              }
              OPTIONAL {
                ?id givenName ?givenName . 
              }
              OPTIONAL {
                ?id familyName ?familyName .
              }
              {
                SELECT * WHERE {
                  { ?id <https://bluebrain.github.io/nexus/vocabulary/deprecated> "false"^^xsd:boolean ; a Class ; 
 path_1 ?v0 . 
 FILTER(?v0 = "value_1") ; name ?name  FILTER (?name = "EPFL") } UNION
                  { ?id <https://bluebrain.github.io/nexus/vocabulary/deprecated> "false"^^xsd:boolean ; a Class ; 
 path_1 ?v0 . 
 FILTER(?v0 = "value_1") ; familyName ?familyName; givenName ?givenName  FILTER (?givenName = "EPFL") } UNION
                  { ?id <https://bluebrain.github.io/nexus/vocabulary/deprecated> "false"^^xsd:boolean ; a Class ; 
 path_1 ?v0 . 
 FILTER(?v0 = "value_1") ; familyName ?familyName; givenName ?givenName  FILTER (?familyName = "EPFL") }
                } LIMIT 1
              }
            }
            """),
                id="agent_query_str"
            )
        ]
    )
def test_write_query(bbn_deprecated_property,filters, query_template, properties_to_filter_with, text, expected_query, model_context):
    constructed_query, limit = _build_resolving_query(text, query_template, bbn_deprecated_property, filters, ResolvingStrategy.EXACT_MATCH, "Class", properties_to_filter_with, model_context, SPARQLQueryBuilder, 10)
    assert limit == 1
    assert expected_query == constructed_query



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union, Tuple

from kgforge.core import Resource
from kgforge.core.commons.attributes import repr_class
from kgforge.core.commons.exceptions import ConfigurationError, ResolvingError
from kgforge.core.commons.execution import not_supported
from kgforge.core.commons.imports import import_class
from kgforge.core.commons.query_builder import QueryBuilder
from kgforge.core.commons.strategies import ResolvingStrategy
from kgforge.core.wrappings.paths import Filter, FilterOperator

class Resolver(ABC):

    # See demo_resolver.py in kgforge/specializations/resolvers/ for a reference implementation.

    # POLICY Methods of archetypes, except __init__, should not have optional arguments.

    # POLICY Implementations should be declared in kgforge/specializations/resolvers/__init__.py.
    # POLICY Implementations should not add methods but private functions in the file.
    # TODO Create a generic parameterizable test suite for resolvers. DKE-135.
    # POLICY Implementations should pass tests/specializations/resolvers/test_resolvers.py.

    def __init__(self, source: str, targets: List[Dict[str, Any]], result_resource_mapping: str,
                 **source_config) -> None:
        # POLICY Resolver data access should be lazy, unless it takes less than a second.
        # POLICY There could be data caching but it should be aware of changes made in the source.
        self.source: str = source
        self.targets: Dict[str, Tuple[str, Dict[str, str]]] = {}
        for target in targets:
            if "filters" in target:
                # reshape filters to match query filters
                filters = {f["path"]: f["value"] for f in target["filters"]}
            else:
                filters = None
            self.targets[target["identifier"]] = {"bucket": target["bucket"], "filters": filters}
        self.result_mapping: Any = self.mapping.load(result_resource_mapping)
        self.service: Any = self._initialize_service(self.source, self.targets, **source_config)

    def __repr__(self) -> str:
        return repr_class(self)

    @property
    @abstractmethod
    def mapping(self) -> Callable:
        """Mapping class to load result_resource_mapping."""
        pass

    @property
    @abstractmethod
    def mapper(self) -> Callable:
        """Mapper class to map the result data to a Resource with result_resource_mapping."""
        pass

    def resolve(self, text: Union[str, List[str], Resource], target: str, type: str,
                strategy: ResolvingStrategy, resolving_context: Any, property_to_resolve: str, merge_inplace_as: str,
                limit: int, threshold: float) \
            -> Optional[Union[Resource, List[Resource], Dict[str, List[Resource]]]]:

        if isinstance(text, Resource):
            expected = [isinstance(text, Resource), property_to_resolve is not None and hasattr(text, property_to_resolve),
                        isinstance(getattr(text, property_to_resolve), str) or
                        (isinstance(getattr(text, property_to_resolve), list) and all([isinstance(f, str) for f in
                                                                                       getattr(text, property_to_resolve)
                                                                                       ]))]
            if all(expected):
                text_to_resolve = getattr(text, property_to_resolve)
            else:
                raise ResolvingError(
                    "When resolving a Resource, a property_to_resolve of type str or List[str] should be provided")

        elif property_to_resolve is not None or merge_inplace_as is not None:
            not_supported(("property_to_resolve or merge_inplace_as", str))
        else:
            text_to_resolve = text
        # The resolving strategy cannot be abstracted as it should be managed by the service.
        self._is_target_valid(target)
        resolved = self._resolve(text_to_resolve, target, type, strategy, resolving_context, limit, threshold)
        if resolved is None or len(resolved) == 0:
            return None
        resolved = resolved[0] if len(resolved) == 1 else resolved
        if isinstance(resolved, tuple):
            # Case Tuple[str,List[Dict]]
            resolved_mapped = self.mapper().map(resolved[1], self.result_mapping, None) if resolved[1] is not None else None
        elif isinstance(text_to_resolve, list):
            # Case List[Tuple[str, List[Dict]]]
            resolved_mapped = {r[0]: self.mapper().map(r[1], self.result_mapping, None) for r in resolved if
                               isinstance(r, tuple)}
        else:
            # Case Dict or List[Dict]
            resolved_mapped = self.mapper().map(resolved, self.result_mapping, None)
        if isinstance(text, Resource) and isinstance(merge_inplace_as, str):
            text.__setattr__(merge_inplace_as, resolved_mapped)
            return text
        else:
            return resolved_mapped

    @abstractmethod
    def _resolve(self, text: Union[str, List[str], Resource], target: str, type: str,
                 strategy: ResolvingStrategy, resolving_context: Any, limit: int, threshold: float) -> Optional[
        List[Any]]:
        # POLICY Should notify of failures with exception ResolvingError including a message.
        pass

    @abstractmethod
    def _is_target_valid(self, target: str) -> Optional[bool]:
        # POLICY Should notify of failures with exception ValueError including a message.
        pass


    # Utils.

    def _initialize_service(self, source: str, targets: Dict[str, Dict[str, Dict[str, str]]], **source_config) -> Any:
        # Resolver data could be accessed from a directory, a web service, or a Store.
        # Initialize the access to the resolver data according to the source type.
        # POLICY Should not use 'self'. This is not a function only for the specialization to work.
        origin = source_config.pop("origin")
        if origin == "directory":
            dirpath = Path(source)
            return self._service_from_directory(dirpath, targets, **source_config)
        elif origin == "web_service":
            return self._service_from_web_service(source, targets)
        elif origin == "store":
            store = import_class(source, "stores")
            return self._service_from_store(store, targets, **source_config)
        else:
            raise ConfigurationError(f"unrecognized Resolver origin '{origin}'")

    @staticmethod
    @abstractmethod
    def _service_from_directory(dirpath: Path, targets: Dict[str, Dict[str, Dict[str, str]]], **source_config) -> Any:
        pass

    @staticmethod
    def _service_from_web_service(endpoint: str, targets: Dict[str,  Dict[str, Dict[str, str]]]) -> Any:
        not_supported()

    @staticmethod
    def _service_from_store(store: Callable, targets: Dict[str,  Dict[str, Dict[str, str]]], **store_config) -> Any:
        not_supported()


def escape_punctuation(text):
    if not isinstance(text, str):
        raise TypeError('Only accepting strings.')
    punctuation = "-()\"#/@;:<>{}`+=~|.!?,"
    for p in punctuation:
        if p in text:
            text = text.replace(p, f"\\\\{p}")
    return text

def write_sparql_filters(text, properties: List, regex=False,
                         case_insensitive=False) -> List[str]:
    if case_insensitive and not regex:
        raise ValueError('Case-insensitive option only valid when regex is True.')
    start_str = " FILTER regex(" if regex else " FILTER ("
    end_str = ", \"i\")" if case_insensitive else ")"
    filters = []
    for property in properties:
        if regex:
            full_str = start_str + f"?{property}, \"{text}\"" + end_str
        else:
            full_str = start_str + f"?{property} = \"{text}\"" + end_str
        filters.append(full_str)
    return filters


def _build_resolving_query(text, query_template, deprecated_property, filters, strategy, _type, properties_to_filter_with, resolving_context: Any, query_builder: QueryBuilder, limit: Optional[int]):
    first_filters = f"?id <{deprecated_property}> \"false\"^^xsd:boolean"
    if _type:
        first_filters = f"{first_filters} ; a {_type}"

    if strategy == strategy.EXACT_MATCH:
        regex = False
        case_insensitive = False
        limit = 1
    elif strategy == strategy.EXACT_CASEINSENSITIVE_MATCH:
        regex = True
        case_insensitive = True
        text = f"^{escape_punctuation(text)}$"
        limit = 1
    else:
        regex = True
        case_insensitive = True
        if strategy == strategy.BEST_MATCH:
            limit = 1

    properties_filters = write_sparql_filters(text, properties_to_filter_with,
                                            regex, case_insensitive)
    
    configured_target_filters = []
    if filters:
        for path, value in filters.items():
            path = path.split(".") if '.' in path else [path]
            configured_target_filters.append(Filter(operator=FilterOperator.EQUAL, path=path, value=value))
        target_query_statements, target_query_filters = query_builder.build(None,
                                                            None,
                                                            resolving_context,
                                                            *configured_target_filters)
        
        target_query_statements = ";\n ".join(target_query_statements)
        target_query_filters = "\n ".join(target_query_filters)
        first_filters = f"{first_filters} ; \n {target_query_statements}"
        first_filters = f"{first_filters} . \n {target_query_filters}" if len(target_query_filters) > 0 else first_filters
    query = query_template.format(first_filters, *properties_filters, limit)
    return query, limit



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
from enum import Enum, auto

from typing import Any, Dict, List, Union

from kgforge.core.commons.attributes import check_collisions, repr_class


class FilterOperator(Enum):
    EQUAL = "__eq__"
    NOT_EQUAL = "__ne__"
    LOWER_THAN = "__lt__"
    LOWER_OR_Equal_Than = "__le__"
    GREATER_Than = "__gt__"
    GREATER_OR_Equal_Than = "__ge__"


class Filter:
    def __init__(self, path: List[str], operator: str, value: Any) -> None:
        self.path: List[str] = path
        try:
            self.operator: str = FilterOperator(operator).value
        except Exception as e:
            raise ValueError(
                f"Invalid operator value '{operator}'. Allowed operators are {[member.value for name, member in FilterOperator.__members__.items()]}"
            )
        self.value: Any = value

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.path == other.path
            and self.operator == other.operator
            and self.value == other.value
        )

    def __repr__(self) -> str:
        return repr_class(self)


class FilterMixin:

    _RESERVED = {"_path"}

    def __init__(self, path: List[str]) -> None:
        self._path: List[str] = path

    def __repr__(self) -> str:
        return repr_class(self)

    def __lt__(self, other: Any) -> Filter:
        return self._for(FilterOperator.LOWER_THAN.value, other)

    def __le__(self, other: Any) -> Filter:
        return self._for(FilterOperator.LOWER_OR_Equal_Than.value, other)

    def __eq__(self, other: Any) -> Filter:
        return self._for(FilterOperator.EQUAL.value, other)

    def __ne__(self, other: Any) -> Filter:
        return self._for(FilterOperator.NOT_EQUAL.value, other)

    def __gt__(self, other: Any) -> Filter:
        return self._for(FilterOperator.GREATER_Than.value, other)

    def __ge__(self, other: Any) -> Filter:
        return self._for(FilterOperator.GREATER_OR_Equal_Than.value, other)

    def _for(self, operator: str, other: Any) -> Filter:
        return Filter(self._path, operator, other)


class PathWrapper(FilterMixin):
    def __init__(self, path: List[str]) -> None:
        super().__init__(path)


class PathsWrapper(FilterMixin):
    def __init__(self, path: List[str], paths: Dict) -> None:
        check_collisions(self._RESERVED, paths.keys())
        super().__init__(path)
        self.__dict__ = paths


def wrap_paths(template: Dict) -> PathsWrapper:
    return _wrap(template, [])


def _wrap(data: Any, path: List[str]) -> Union[PathsWrapper, PathWrapper]:
    if isinstance(data, Dict):
        return PathsWrapper(path, {k: _wrap(v, path + [k]) for k, v in data.items()})
    else:
        return PathWrapper(path)


def create_filters_from_dict(filter_dict: Dict, path_prefix=None) -> List[Filter]:
    filters = list()
    if path_prefix is None:
        path_prefix = []
    for k, v in filter_dict.items():
        if '/' in k:
            path = path_prefix + k.split('/')
        else:
            path = path_prefix + [k]
        if isinstance(v, dict):
            filters.extend(create_filters_from_dict(v, path))
        else:
            filters.append(Filter(path, FilterOperator.EQUAL.value, v))
    return filters



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
import datetime

import asyncio
import copy

import json
import mimetypes
import re
from asyncio import Semaphore, Task
from enum import Enum
from json import JSONDecodeError

from kgforge.core.commons.dictionaries import update_dict
from kgforge.core.commons.es_query_builder import ESQueryBuilder
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from urllib.parse import quote_plus, unquote, urlparse, parse_qs
from kgforge.core.commons.sparql_query_builder import SPARQLQueryBuilder

import nexussdk as nexus
import requests
from aiohttp import ClientSession, MultipartWriter
from aiohttp.hdrs import CONTENT_DISPOSITION, CONTENT_TYPE
from numpy import nan
from pyld import jsonld
from rdflib import Graph
from rdflib.plugins.sparql.parser import Query
from datetime import datetime
from requests import HTTPError

from kgforge.core import Resource
from kgforge.core.archetypes import Store
from kgforge.core.commons.actions import LazyAction
from kgforge.core.commons.context import Context
from kgforge.core.commons.exceptions import (
    DeprecationError,
    DownloadingError,
    QueryingError,
    RegistrationError,
    RetrievalError,
    TaggingError,
    UpdatingError,
    UploadingError,
)
from kgforge.core.commons.execution import run, not_supported
from kgforge.core.commons.files import is_valid_url
from kgforge.core.commons.parser import _parse_type
from kgforge.core.conversions.json import as_json
from kgforge.core.conversions.rdf import as_jsonld, from_jsonld
from kgforge.core.wrappings.dict import DictWrapper
from kgforge.core.wrappings.paths import Filter, create_filters_from_dict
from kgforge.specializations.mappers import DictionaryMapper
from kgforge.specializations.mappings import DictionaryMapping
from kgforge.specializations.stores.nexus.service import BatchAction, Service


class CategoryDataType(Enum):
    DATETIME = "datetime"
    NUMBER = "number"
    BOOLEAN = "boolean"
    LITERAL = "literal"


type_map = {
    datetime: CategoryDataType.DATETIME,
    str: CategoryDataType.LITERAL,
    bool: CategoryDataType.BOOLEAN,
    int: CategoryDataType.NUMBER,
    float: CategoryDataType.NUMBER,
    complex: CategoryDataType.NUMBER,
}

format_type = {
    CategoryDataType.DATETIME: lambda x: f'"{x}"^^xsd:dateTime',
    CategoryDataType.NUMBER: lambda x: x,
    CategoryDataType.LITERAL: lambda x: f'"{x}"',
    CategoryDataType.BOOLEAN: lambda x: "'true'^^xsd:boolean" if x is True else "'false'^^xsd:boolean",
}

sparql_operator_map = {
    "__lt__": "<",
    "__le__": "<=",
    "__eq__": "=",
    "__ne__": "!=",
    "__gt__": ">",
    "__ge__": ">=",
}

elasticsearch_operator_map = {
    "__lt__": "lt",
    "__le__": "lte",
    "__gt__": "gt",
    "__ge__": "gte",
}


class BlueBrainNexus(Store):
    def __init__(
        self,
        endpoint: Optional[str] = None,
        bucket: Optional[str] = None,
        token: Optional[str] = None,
        versioned_id_template: Optional[str] = None,
        file_resource_mapping: Optional[str] = None,
        model_context: Optional[Context] = None,
        searchendpoints: Optional[Dict] = None,
        **store_config,
    ) -> None:
        super().__init__(
            endpoint,
            bucket,
            token,
            versioned_id_template,
            file_resource_mapping,
            model_context,
            searchendpoints,
            **store_config,
        )

    @property
    def mapping(self) -> Optional[Callable]:
        return DictionaryMapping

    @property
    def mapper(self) -> Optional[DictionaryMapper]:
        return DictionaryMapper
    
    def register(
        self, data: Union[Resource, List[Resource]], schema_id: str = None
    ) -> None:
        run(
            self._register_one,
            self._register_many,
            data,
            required_synchronized=False,
            execute_actions=True,
            exception=RegistrationError,
            monitored_status="_synchronized",
            schema_id=schema_id,
        )

    def _register_many(self, resources: List[Resource], schema_id: str) -> None:
        def register_callback(task: Task):
            result = task.result()
            if isinstance(result.response, Exception):
                self.service.synchronize_resource(
                    result.resource,
                    result.response,
                    self._register_many.__name__,
                    False,
                    False,
                )
            else:
                result.resource.id = result.response["@id"]
                if not hasattr(result.resource, "context"):
                    context = self.model_context or self.context
                    result.resource.context = (
                        context.iri
                        if context.is_http_iri()
                        else context.document["@context"]
                    )
                self.service.synchronize_resource(
                    result.resource,
                    result.response,
                    self._register_many.__name__,
                    True,
                    True,
                )

        verified = self.service.verify(
            resources,
            self._register_many.__name__,
            RegistrationError,
            id_required=False,
            required_synchronized=False,
            execute_actions=True,
        )
        params_register = copy.deepcopy(self.service.params.get("register", {}))
        self.service.batch_request(
            verified,
            BatchAction.CREATE,
            register_callback,
            RegistrationError,
            schema_id=schema_id,
            params=params_register,
        )

    def _register_one(self, resource: Resource, schema_id: str) -> None:
        context = self.model_context or self.context
        data = as_jsonld(
            resource,
            "compacted",
            False,
            model_context=context,
            metadata_context=None,
            context_resolver=self.service.resolve_context
        )

        try:
            schema = quote_plus(schema_id) if schema_id else "_"
            url_base = f"{self.service.url_resources}/{schema}"
            params_register = copy.deepcopy(self.service.params.get("register", None))
            identifier = resource.get_identifier()
            if identifier:
                url = f"{url_base}/{quote_plus(identifier)}"
                response = requests.put(
                    url,
                    headers=self.service.headers,
                    data=json.dumps(data, ensure_ascii=True),
                    params=params_register,
                )
            else:
                url = url_base
                response = requests.post(
                    url,
                    headers=self.service.headers,
                    data=json.dumps(data, ensure_ascii=True),
                    params=params_register,
                )
            response.raise_for_status()

        except nexus.HTTPError as e:
            raise RegistrationError(_error_message(e))
        else:
            response_json = response.json()
            resource.id = response_json["@id"]
            # If resource had no context, update it with the one provided by the store.
            if not hasattr(resource, "context"):
                resource.context = data["@context"]
            self.service.sync_metadata(resource, response_json)

    def _upload_many(self, paths: List[Path], content_type: str) -> List[Dict]:
        async def _bulk():
            loop = asyncio.get_event_loop()
            semaphore = Semaphore(self.service.max_connection)
            async with ClientSession(headers=self.service.headers_upload) as session:
                tasks = (_create_task(x, loop, semaphore, session) for x in paths)
                return await asyncio.gather(*tasks)

        def _create_task(path, loop, semaphore, session):
            default = "application/octet-stream"
            mime_type = content_type or mimetypes.guess_type(str(path))[0] or default
            # FIXME Nexus seems to not parse the Content-Disposition 'filename*' field  properly.
            # data = FormData()
            # data.add_field("file", path.open("rb"), content_type=mime_type, filename=path.name)
            # FIXME This hack is to prevent sending Content-Disposition with the 'filename*' field.
            data = MultipartWriter("form-data")
            part = data.append(path.open("rb"))
            part.headers[CONTENT_TYPE] = mime_type
            part.headers[
                CONTENT_DISPOSITION
            ] = f'form-data; name="file"; filename="{path.name}"'
            return loop.create_task(_upload(data, semaphore, session))

        async def _upload(data, semaphore, session):
            async with semaphore:
                async with session.post(self.service.url_files, data=data) as response:
                    body = await response.json()
                    if response.status < 400:
                        return body
                    else:
                        msg = " ".join(
                            re.findall("[A-Z][^A-Z]*", body["@type"])
                        ).lower()
                        raise UploadingError(msg)

        return asyncio.run(_bulk())

    def _upload_one(self, path: Path, content_type: str) -> Dict:
        file = str(path.absolute())
        mime_type = content_type or mimetypes.guess_type(file, True)[0]
        if mime_type is None:
            mime_type = "application/octet-stream"
        try:
            response = nexus.files.create(
                self.organisation, self.project, file, content_type=mime_type
            )
        except HTTPError as e:
            raise UploadingError(_error_message(e))
        else:
            return response

    # C[R]UD.

    def retrieve(
        self, id: str, version: Optional[Union[int, str]], cross_bucket: bool, **params
    ) -> Resource:
        """
        Retrieve a resource by its identifier from the configured store and possibly at a given version.

        :param id: the resource identifier to retrieve
        :param version: a version of the resource to retrieve
        :param cross_bucket: instructs the configured store to whether search beyond the configured bucket (True) or not (False)
        :param params: a dictionary of parameters. Supported parameters are:
              [retrieve_source] whether to retrieve the resource payload as registered in the last update
              (default: True)
        :return: Resource
        """
        version_params = None
        if version is not None:
            if isinstance(version, int):
                version_params = {"rev": version}
            elif isinstance(version, str):
                version_params = {"tag": version}
            else:
                raise RetrievalError("incorrect 'version'")
        parsed_id = urlparse(id)
        fragment = None
        query_params = None
        # urlparse is not separating fragment and query params when the latter are put after a fragment
        if parsed_id.fragment is not None and "?" in str(parsed_id.fragment):
            fragment_parts = urlparse(parsed_id.fragment)
            query_params = parse_qs(fragment_parts.query)
            fragment = fragment_parts.path
        elif parsed_id.fragment is not None and parsed_id.fragment != "":
            fragment = parsed_id.fragment
        elif parsed_id.query is not None and parsed_id.query != "":
            query_params = parse_qs(parsed_id.query)

        if version_params is not None:
            if not isinstance(query_params, dict):
                query_params = {}
            query_params.update(version_params)

        id_without_query = f"{parsed_id.scheme}://{parsed_id.netloc}{parsed_id.path}{'#' + fragment if fragment is not None else ''}"
        url_base = (
            self.service.url_resolver if cross_bucket else self.service.url_resources
        )
        url_resource = "/".join((url_base, "_", quote_plus(id_without_query)))
        retrieve_source = params.get('retrieve_source', True)

        if retrieve_source and not cross_bucket:
            url_source = "/".join((url_resource, "source"))
            url = url_source
        else:
            url = url_resource
        try:
            response = requests.get(
                url, params=query_params, headers=self.service.headers
            )
            response.raise_for_status()
        except HTTPError as er:
            if cross_bucket:
                nexus_path = f"{self.service.endpoint}/resources/" 
            else:
                nexus_path = self.service.url_resources
            # Try to use the id as it was given
            if id.startswith(nexus_path):
                url_resource = id_without_query
                if retrieve_source and not cross_bucket:
                    url = "/".join((id_without_query, "source"))
                else: 
                    url = id_without_query
                try:
                    response = requests.get(
                        url, params=query_params, headers=self.service.headers
                    )
                    response.raise_for_status()
                except HTTPError as e:
                    raise RetrievalError(_error_message(e))
            else:
                raise RetrievalError(_error_message(er))
        finally:
            if retrieve_source and not cross_bucket:

                response_metadata = requests.get(
                    url_resource, params=query_params, headers=self.service.headers
                )
                response_metadata.raise_for_status()
            elif retrieve_source and cross_bucket and response and ('_self' in response.json()):
                response_metadata = requests.get(
                    "/".join([response.json()["_self"], "source"]), params=query_params, headers=self.service.headers
                )
                response_metadata.raise_for_status()
            else:
                response_metadata = True  # when retrieve_source is False
        if response and response_metadata:
            try:
                data = response.json()
                resource = self.service.to_resource(data)
            except Exception as e:
                raise ValueError(e)

            try:
                if retrieve_source and not cross_bucket:
                    data = response_metadata.json()
                if retrieve_source and cross_bucket:
                    resource = self.service.to_resource(response_metadata.json())
            except Exception as e:
                self.service.synchronize_resource(
                        resource, data, self.retrieve.__name__, False, False
                )
                raise ValueError(e)


            finally:
                self.service.synchronize_resource(
                    resource, data, self.retrieve.__name__, True, True
                )
            return resource

    def _retrieve_filename(self, id: str) -> Tuple[str, str]:
        try:
            response = requests.get(id, headers=self.service.headers)
            response.raise_for_status()
            metadata = response.json()
            return metadata["_filename"], metadata["_mediaType"]
        except HTTPError as e:
            raise DownloadingError(_error_message(e))

    def _download_many(
        self,
        urls: List[str],
        paths: List[str],
        store_metadata: Optional[DictWrapper],
        cross_bucket: bool,
        content_type: str,
        buckets: List[str]
    ) -> None:
        async def _bulk():
            loop = asyncio.get_event_loop()
            semaphore = Semaphore(self.service.max_connection)
            headers = self.service.headers_download if not content_type else update_dict(self.service.headers_download, {"Accept":content_type})
            async with ClientSession(headers=headers) as session:
                tasks = (
                    _create_task(x, y, z, b, loop, semaphore, session)
                    for x, y, z, b in zip(urls, paths, store_metadata, buckets)
                )
                return await asyncio.gather(*tasks)

        def _create_task(url, path, store_metadata, bucket, loop, semaphore, session):
            return loop.create_task(
                _download(url, path, store_metadata, bucket, semaphore, session)
            )

        async def _download(url, path, store_metadata, bucket, semaphore, session):
            async with semaphore:
                params_download = copy.deepcopy(self.service.params.get("download", {}))
                async with session.get(url, params=params_download) as response:
                    try:
                        response.raise_for_status()
                    except Exception as e:
                        raise DownloadingError(
                            f"Downloading url {url} from bucket {bucket} failed: {_error_message(e)}"
                        )
                    else:
                        with open(path, "wb") as f:
                            data = await response.read()
                            f.write(data)

        return asyncio.run(_bulk())

    def _download_one(
        self,
        url: str,
        path: str,
        store_metadata: Optional[DictWrapper],
        cross_bucket: bool,
        content_type: str,
        bucket: str
    ) -> None:
        
        try:
            params_download = copy.deepcopy(self.service.params.get("download", {}))
            headers = self.service.headers_download if not content_type else update_dict(self.service.headers_download, {"Accept": content_type})

            response = requests.get(
                url=url,
                headers=headers,
                params=params_download
            )
            response.raise_for_status()
        except Exception as e:
            raise DownloadingError(
                f"Downloading from bucket {bucket} failed: {_error_message(e)}"
            )
        else:
            with open(path, "wb") as f:
                for chunk in response.iter_content(chunk_size=4096):
                    f.write(chunk)

    def _prepare_download_one(
        self,
        url: str,
        store_metadata: Optional[DictWrapper],
        cross_bucket: bool
    ) -> Tuple[str, str, str]:
        if cross_bucket:
            if store_metadata is not None:
                project = store_metadata._project.split("/")[-1]
                org = store_metadata._project.split("/")[-2]
            else:
                raise ValueError(
                    f"Downloading non registered file is not allowed when cross_bucket is set to {cross_bucket}"
                )
        else:
            org = self.service.organisation
            project = self.service.project
        file_id = url.split("/")[-1]
        file_id = unquote(file_id)
        if len(file_id) < 1:
            raise DownloadingError(f"Invalid file url: {url}")
        elif file_id.startswith("http"):
            url_base = url
        else: 
            # this is a hack since _self and _id have the same uuid
            url_base = "/".join(
                (
                    self.service.url_base_files,
                    quote_plus(org),
                    quote_plus(project),
                    quote_plus(file_id),
                )
            )
        return url_base, f"{org}/{project}"

    # CR[U]D.

    def update(self, data: Union[Resource, List[Resource]], schema_id: str) -> None:
        run(
            self._update_one,
            self._update_many,
            data,
            id_required=True,
            required_synchronized=False,
            execute_actions=True,
            exception=UpdatingError,
            monitored_status="_synchronized",
            schema_id=schema_id,
        )

    def _update_many(self, resources: List[Resource], schema_id: str) -> None:
        update_callback = self.service.default_callback(self._update_many.__name__)
        verified = self.service.verify(
            resources,
            self._update_many.__name__,
            UpdatingError,
            id_required=True,
            required_synchronized=False,
            execute_actions=True,
        )
        params_update = copy.deepcopy(self.service.params.get("update", {}))
        self.service.batch_request(
            verified,
            BatchAction.UPDATE,
            update_callback,
            UpdatingError,
            params=params_update,
        )

    def _update_one(self, resource: Resource, schema_id: str) -> None:
        context = self.model_context or self.context
        data = as_jsonld(
            resource,
            "compacted",
            False,
            model_context=context,
            metadata_context=None,
            context_resolver=self.service.resolve_context
        )
        params_update = copy.deepcopy(self.service.params.get("update", {}))
        params_update["rev"] = resource._store_metadata._rev
        schema = quote_plus(schema_id) if schema_id else "_"
        url = f"{self.service.url_resources}/{schema}/{quote_plus(resource.id)}"
        try:
            response = requests.put(
                url,
                headers=self.service.headers,
                data=json.dumps(data, ensure_ascii=True),
                params=params_update,
            )
            response.raise_for_status()
        except HTTPError as e:
            raise UpdatingError(_error_message(e))
        else:
            self.service.sync_metadata(resource, response.json())

    def tag(self, data: Union[Resource, List[Resource]], value: str) -> None:
        run(
            self._tag_one,
            self._tag_many,
            data,
            id_required=True,
            required_synchronized=True,
            exception=TaggingError,
            value=value,
        )

    def _tag_many(self, resources: List[Resource], value: str) -> None:
        tag_callback = self.service.default_callback(self._tag_many.__name__)
        verified = self.service.verify(
            resources,
            self._tag_many.__name__,
            TaggingError,
            id_required=True,
            required_synchronized=True,
            execute_actions=False,
        )
        params_tag = copy.deepcopy(self.service.params.get("tag", {}))
        self.service.batch_request(
            verified,
            BatchAction.TAG,
            tag_callback,
            TaggingError,
            tag=value,
            params=params_tag,
        )

    def _tag_one(self, resource: Resource, value: str) -> None:
        data = {"tag": value, "rev": resource._store_metadata._rev}
        url = f"{self.service.url_resources}/_/{quote_plus(resource.id)}/tags?rev={resource._store_metadata._rev}"
        try:
            params_tag = copy.deepcopy(self.service.params.get("tag", None))
            response = requests.post(
                url,
                headers=self.service.headers,
                data=json.dumps(data, ensure_ascii=True),
                params=params_tag,
            )
            response.raise_for_status()
        except HTTPError as e:
            raise TaggingError(_error_message(e))
        else:
            self.service.sync_metadata(resource, response.json())

    # CRU[D].

    def deprecate(self, data: Union[Resource, List[Resource]]) -> None:
        run(
            self._deprecate_one,
            self._deprecate_many,
            data,
            id_required=True,
            required_synchronized=True,
            exception=DeprecationError,
            monitored_status="_synchronized",
        )

    def _deprecate_many(self, resources: List[Resource]) -> None:
        deprecate_callback = self.service.default_callback(
            self._deprecate_many.__name__
        )
        verified = self.service.verify(
            resources,
            self._deprecate_many.__name__,
            DeprecationError,
            id_required=True,
            required_synchronized=True,
            execute_actions=False,
        )
        params_deprecate = copy.deepcopy(self.service.params.get("deprecate", {}))
        self.service.batch_request(
            verified,
            BatchAction.DEPRECATE,
            deprecate_callback,
            DeprecationError,
            params=params_deprecate,
        )

    def _deprecate_one(self, resource: Resource) -> None:
        url = f"{self.service.url_resources}/_/{quote_plus(resource.id)}?rev={resource._store_metadata._rev}"
        try:
            params_deprecate = copy.deepcopy(self.service.params.get("deprecate", None))
            response = requests.delete(
                url, headers=self.service.headers, params=params_deprecate
            )
            response.raise_for_status()
        except HTTPError as e:
            raise DeprecationError(_error_message(e))
        else:
            self.service.sync_metadata(resource, response.json())

        # Querying.

    def search(
        self, resolvers: Optional[List["Resolver"]], *filters, **params
    ) -> List[Resource]:

        if self.model_context is None:
            raise ValueError("context model missing")

        debug = params.get("debug", False)
        limit = params.get("limit", 100)
        offset = params.get("offset", None)
        deprecated = params.get("deprecated", False)
        cross_bucket = params.get("cross_bucket", False)
        bucket = params.get("bucket", None)
        search_in_graph = params.get("search_in_graph", True)
        distinct = params.get("distinct", False)
        includes = params.get("includes", None)
        excludes = params.get("excludes", None)
        search_endpoint = params.get(
            "search_endpoint", self.service.sparql_endpoint["type"]
        )
        if search_endpoint not in [
            self.service.sparql_endpoint["type"],
            self.service.elastic_endpoint["type"],
        ]:
            raise ValueError(
                f"The provided search_endpoint value '{search_endpoint}' is not supported. Supported "
                f"search_endpoint values are: '{self.service.sparql_endpoint['type'], self.service.elastic_endpoint['type']}'"
            )
        if "filters" in params:
            raise ValueError(
                "A 'filters' key was provided as params. Filters should be provided as iterable to be unpacked.")

        if bucket and not cross_bucket:
            not_supported(("bucket", True))

        if filters and isinstance(filters[0], dict):
            filters = create_filters_from_dict(filters[0])
        filters = list(filters) if not isinstance(filters, list) else filters

        if search_endpoint == self.service.sparql_endpoint["type"]:
            if includes or excludes:
                raise ValueError(
                    "Field inclusion and exclusion are not supported when using SPARQL"
                )
            project_filter = ""
            if bucket:
                project_filter = f"Filter (?_project = <{'/'.join([self.endpoint, 'projects', bucket])}>)"
            elif not cross_bucket:
                project_filter = f"Filter (?_project = <{'/'.join([self.endpoint, 'projects', self.organisation, self.project])}>)"

            query_statements, query_filters = SPARQLQueryBuilder.build(
                None, resolvers, self.model_context, *filters
            )
            retrieve_source = params.get("retrieve_source", True)
            store_metadata_statements = []
            if retrieve_source:
                _vars = ["?id"]
                for i, k in enumerate(self.service.store_metadata_keys):
                    _vars.append(f"?{k}")
                    store_metadata_statements.insert(
                        i + 2, f"<{self.metadata_context.terms[k].id}> ?{k}")
                deprecated_filter = f"Filter (?_deprecated = {format_type[CategoryDataType.BOOLEAN](deprecated)})"
                query_filters.append(deprecated_filter)
            else:
                _vars = ["?id", "?_project", "?_rev"]
                store_metadata_statements.append(f"<{self.service.revision_property}> ?_rev")
                store_metadata_statements.append(f"<{self.service.project_property}> ?_project")
                query_statements.append(
                    f"<{self.service.deprecated_property}> {format_type[CategoryDataType.BOOLEAN](deprecated)}",
                )
            query_statements.extend(store_metadata_statements)
            statements = ";\n ".join(query_statements)
            _filters = "\n".join(
                (".\n ".join(query_filters), project_filter)
            )
            query = _create_select_query(
                _vars, f"?id {statements} . \n {_filters}", distinct,
                search_in_graph
            )
            # support @id and @type
            resources = self.sparql(query, debug=debug, limit=limit, offset=offset)
            params_retrieve = copy.deepcopy(self.service.params.get("retrieve", {}))
            params_retrieve['retrieve_source'] = retrieve_source
            results = self.service.batch_request(
                resources, BatchAction.FETCH, None, QueryingError, params=params_retrieve
            )
            resources = list()
            for result in results:
                resource = result.resource
                if retrieve_source:
                    store_metadata_response = as_json(result.resource, expanded=False, store_metadata=False,
                                                      model_context=None,
                                                      metadata_context=None,
                                                      context_resolver=None)  # store_metadata is obtained from SPARQL (resource) and not from server (response) because of retrieve_source==True
                else:
                    store_metadata_response = result.response  # dict
                try:
                    resource = self.service.to_resource(result.response)
                except Exception as e:
                    self.service.synchronize_resource(
                        resource, store_metadata_response, self.search.__name__, False, False
                    )
                    raise ValueError(e)
                finally:
                    self.service.synchronize_resource(
                        resource, store_metadata_response, self.search.__name__, True, False
                    )
                resources.append(resource)
            return resources
        else:
            if isinstance(self.service.elastic_endpoint["view"], LazyAction):
                self.service.elastic_endpoint["view"] = self.service.elastic_endpoint[
                    "view"
                ].execute()

            elastic_mapping = (
                self.service.elastic_endpoint["view"]["mapping"]
                if "mapping" in self.service.elastic_endpoint["view"]
                else None
            )
            default_str_keyword_field = self.service.elastic_endpoint[
                "default_str_keyword_field"
            ]
            deprecated_property_context_term = self.service.metadata_context.find_term(
                self.service.deprecated_property
            )
            project_property_context_term = self.service.metadata_context.find_term(
                self.service.project_property
            )
            filters.append(
                Filter(
                    operator="__eq__",
                    path=[
                        deprecated_property_context_term.name
                        if deprecated_property_context_term is not None
                        else "_deprecated"
                    ],
                    value=deprecated,
                )

            )
            _project = None
            if bucket:
                _project = '/'.join([self.endpoint, 'projects', bucket])

            elif not cross_bucket:
                _project = '/'.join([self.endpoint, 'projects', self.organisation, self.project])

            if _project:
                filters.append(
                    Filter(
                        operator="__eq__",
                        path=[
                            project_property_context_term.name
                            if project_property_context_term is not None
                            else "_project"
                        ],
                        value=_project
                    )
                )

            query = ESQueryBuilder.build(
                elastic_mapping,
                resolvers,
                self.model_context,
                filters,
                default_str_keyword_field=default_str_keyword_field,
                includes=includes,
                excludes=excludes,
            )

            return self.elastic(
                json.dumps(query), debug=debug, limit=limit, offset=offset
            )

    def _sparql(self, query: str) -> List[Resource]:
        try:
            response = requests.post(
                self.service.sparql_endpoint["endpoint"],
                data=query,
                headers=self.service.headers_sparql,
            )
            response.raise_for_status()
        except Exception as e:
            raise QueryingError(e)
        else:
            data = response.json()
            # FIXME workaround to parse a CONSTRUCT query, this fix depends on
            #  https://github.com/BlueBrain/nexus/issues/1155
            _, q_comp = Query.parseString(query)
            if q_comp.name == "ConstructQuery":
                subject_triples = {}
                for r in data["results"]["bindings"]:
                    subject = r["subject"]["value"]
                    s = f"<{r['subject']['value']}>"
                    p = f"<{r['predicate']['value']}>"
                    if r["object"]["type"] == "uri":
                        o = f"<{r['object']['value']}>"
                    else:
                        if "datatype" in r["object"]:
                            o = f"\"{r['object']['value']}\"^^{r['object']['datatype']}"
                        else:
                            o = f"\"{r['object']['value']}\""
                    if subject in subject_triples:
                        subject_triples[subject] += f"\n{s} {p} {o} . "
                    else:
                        subject_triples[subject] = f"{s} {p} {o} . "

                def triples_to_resource(iri, triples):
                    graph = Graph().parse(data=triples, format="nt")
                    data_expanded = json.loads(graph.serialize(format="json-ld"))
                    data_expanded = json.loads(graph.serialize(format="json-ld"))
                    frame = {"@id": iri}
                    data_framed = jsonld.frame(data_expanded, frame)
                    context = self.model_context or self.context
                    compacted = jsonld.compact(data_framed, context.document)
                    resource = from_jsonld(compacted)
                    resource.context = (
                        context.iri
                        if context.is_http_iri()
                        else context.document["@context"]
                    )
                    return resource

                return [triples_to_resource(s, t) for s, t in subject_triples.items()]

            else:
                # SELECT QUERY
                results = data["results"]["bindings"]
                return [
                    Resource(**{k: json.loads(str(v["value"]).lower()) if v['type'] == 'literal' and
                                ('datatype' in v and v['datatype'] ==
                                 'http://www.w3.org/2001/XMLSchema#boolean')
                                else (int(v["value"]) if v['type'] == 'literal' and
                                      ('datatype' in v and v['datatype'] ==
                                       'http://www.w3.org/2001/XMLSchema#integer')
                                      else v["value"]
                                      )
                                for k, v in x.items()})
                    for x in results
                ]

    def _elastic(self, query: str) -> List[Resource]:
        try:
            response = requests.post(
                self.service.elastic_endpoint["endpoint"],
                data=query,
                headers=self.service.headers_elastic,
            )
            response.raise_for_status()
        except Exception as e:
            raise QueryingError(e)
        else:
            results = response.json()
            return [
                self.service.to_resource(
                    hit["_source"],
                    True,
                    **{
                        "id": hit.get("_id", None),
                        "_index": hit.get("_index", None),
                        "_score": hit.get("_score", None),
                    },
                )
                for hit in results["hits"]["hits"]
            ]

    # Utils.

    def _initialize_service(
        self,
        endpoint: Optional[str],
        bucket: Optional[str],
        token: Optional[str],
        searchendpoints: Optional[Dict],
        **store_config,
    ) -> Any:
        try:
            self.organisation, self.project = self.bucket.split("/")
            max_connection = store_config.pop("max_connection", 50)
            if max_connection <= 0:
                raise ValueError(
                    f"max_connection value should be great than 0 but {max_connection} is provided"
                )
            store_context_config = store_config.pop("vocabulary", {})
            nexus_metadata_context = store_context_config.get(
                "metadata",
                {
                    "iri": Service.NEXUS_CONTEXT_FALLBACK,
                    "local_iri": Service.NEXUS_CONTEXT_FALLBACK,
                },
            )
            nexus_context_iri = nexus_metadata_context.get("iri")
            nexus_context_local_iri = nexus_metadata_context.get("local_iri")
            namespace = store_context_config.get(
                "namespace", Service.NEXUS_NAMESPACE_FALLBACK
            )
            project_property = store_context_config.get(
                "project_property", Service.PROJECT_PROPERTY_FALLBACK
            )
            deprecated_property = store_context_config.get(
                "deprecated_property", Service.DEPRECATED_PROPERTY_FALLBACK
            )
            content_type = store_config.pop("Content-Type", "application/ld+json")
            accept = store_config.pop("Accept", "application/ld+json")
            files_upload_config = store_config.pop(
                "files_upload", {"Accept": "application/ld+json"}
            )
            files_download_config = store_config.pop(
                "files_download", {"Accept": "*/*"}
            )
            params = store_config.pop("params", {})
        except Exception as ve:
            raise ValueError(f"Store configuration error: {ve}")
        else:
            return Service(
                endpoint=endpoint,
                org=self.organisation,
                prj=self.project,
                token=token,
                model_context=self.model_context,
                max_connection=max_connection,
                searchendpoints=searchendpoints,
                store_context=nexus_context_iri,
                store_local_context=nexus_context_local_iri,
                namespace=namespace,
                project_property=project_property,
                deprecated_property=deprecated_property,
                content_type=content_type,
                accept=accept,
                files_upload_config=files_upload_config,
                files_download_config=files_download_config,
                **params,
            )
            
    def rewrite_uri(self, uri: str, context: Context, **kwargs) -> str:
        is_file = kwargs.get("is_file", True)
        encoding = kwargs.get("encoding", None)

        # try decoding the url first
        raw_url = unquote(uri)
        if is_file: # for files
            url_base = '/'.join([self.endpoint, 'files', self.bucket])
        else: # for resources
            url_base = '/'.join([self.endpoint, 'resources', self.bucket])
        matches = re.match(r"[\w\.:%/-]+/(\w+):(\w+)/[\w\.-/:%]+", raw_url)
        if matches:
            groups = matches.groups()
            old_schema = f"{groups[0]}:{groups[1]}"
            resolved = context.expand(groups[0])
            if raw_url.startswith(url_base):
                extended_schema = resolved + groups[1]
                url = raw_url.replace(old_schema, quote_plus(extended_schema))
                schema_and_id = url.split(url_base+"/")[1]
                id = schema_and_id.split(quote_plus(extended_schema)+"/")[-1]
                if not is_valid_url(id):        
                    resolved_id = context.resolve_iri(id)
                else:
                    resolved_id = id
                return url.replace(id, quote_plus(resolved_id))
            else:
                extended_schema = ''.join([resolved, groups[1]])
                url = raw_url.replace(old_schema, extended_schema)
        else:
            url = raw_url
        if url.startswith(url_base):
            schema_and_id = url.split(url_base)[1]
            if "/_/" in schema_and_id: # has _ schema
                 id = schema_and_id.split("/_/")[-1]
            else:
                id = schema_and_id.split("/")[-1]
            if not is_valid_url(id):
                resolved_id = context.resolve_iri(id)
            else:
                resolved_id = id
            if resolved_id in schema_and_id:
                return uri # expanded already given
            else:
                return url.replace(id, quote_plus(resolved_id))
        if not is_file and "/_/" not in url: # adding _ for empty schema
            uri = "/".join((url_base, "_", quote_plus(url, encoding=encoding)))
        else:
            uri = "/".join((url_base, quote_plus(url, encoding=encoding)))
        return uri


def _error_message(error: HTTPError) -> str:
    def format_message(msg):
        return "".join([msg[0].lower(), msg[1:-1], msg[-1] if msg[-1] != "." else ""])

    try:
        error_json = error.response.json()
        messages = []
        reason = error_json.get("reason", None)
        details = error_json.get("details", None)
        if reason:
            messages.append(format_message(reason))
        if details:
            messages.append(format_message(details))
        messages = messages if reason or details else [str(error)]
        return ". ".join(messages)
    except Exception as e:
        pass
    try:
        return format_message(error.response.text())
    except Exception:
        return format_message(str(error))


def _create_select_query(vars_, statements, distinct, search_in_graph):
    where_clauses = (
        f"{{ Graph ?g {{{statements}}}}}" if search_in_graph else f"{{{statements}}}"
    )
    join_vars_ = ' '.join(vars_)
    select_vars = f"DISTINCT {join_vars_}" if distinct else f"{join_vars_}"
    return f"SELECT {select_vars} WHERE {where_clauses}"



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
from builtins import all

from pathlib import Path
import requests
from requests import RequestException
from urllib.parse import urlparse

def load_file_as_byte(source: str):
    # source: Union[str, Path, URL].
    filepath = Path(source)
    if filepath.is_file():
        data = filepath.read_bytes()
    else:
        try:
            response = requests.get(source)
            response.raise_for_status()
            data = response.content
        except RequestException as re:
            raise AttributeError(f"Failed to load the configuration from {source}. The provided source is not a valid file path or URL: {str(re)}")
    return data

def is_valid_url(url):
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except Exception as e:
        return False



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
import copy

import datetime
import dateutil
import elasticsearch_dsl
from dateutil.parser import ParserError
from elasticsearch_dsl import Field
from typing import Tuple, List, Dict, Optional, Any

from kgforge.core.commons.context import Context
from kgforge.core.commons.query_builder import QueryBuilder
from kgforge.core.wrappings import Filter, FilterOperator

elasticsearch_operator_range_map = {
    FilterOperator.LOWER_THAN.value: "lt",
    FilterOperator.LOWER_OR_Equal_Than.value: "lte",
    FilterOperator.GREATER_Than.value: "gt",
    FilterOperator.GREATER_OR_Equal_Than.value: "gte",
}


class ESQueryBuilder(QueryBuilder):
    @staticmethod
    def build(
        schema: Dict,
        resolvers: Optional[List["Resolver"]],
        context: Context,
        *filters,
        **params,
    ) -> Tuple[List, List, List]:

        es_filters = list()
        musts = list()
        must_nots = list()
        script_scores = list()

        default_str_keyword_field = params.get("default_str_keyword_field", "keyword")
        includes = params.get("includes", None)
        excludes = params.get("excludes", None)

        m = elasticsearch_dsl.Mapping()
        dynamic = True
        if schema is not None:
            m._update_from_dict(schema)
            dynamic = m._meta["dynamic"] if "dynamic" in m._meta else dynamic

        for index, f in enumerate(*filters):
            _filter = None
            must = None
            must_not = None
            script_score = None
            last_path = f.path[-1]
            property_path = ".".join(f.path)
            found_path_parts = None
            nested_path, mapping_type = m.resolve_nested(field_path=property_path)
            if isinstance(mapping_type, elasticsearch_dsl.Nested) or isinstance(
                mapping_type, elasticsearch_dsl.Object
            ):
                raise ValueError(
                    f"The provided path {f.path} can't be used for filter/search because it is not a leaf property."
                )
            if (
                isinstance(mapping_type, elasticsearch_dsl.DenseVector)
                and f.operator != FilterOperator.EQUAL.value
            ):
                raise ValueError(
                    f"The provided DenseVector path '{f.path}' can't be used for filter/search but only with the "
                    f"'__eq__' operator for similarity search."
                )

            if mapping_type is None and nested_path == ():
                if not dynamic:
                    raise ValueError(
                        f"The provided path {f.path} is not present in the provided elasticsearch mapping (dynamic == {dynamic}). "
                        f"Please search with a path present in the elasticsearch mapping or provide a dynamic mapping."
                    )
                else:
                    (
                        nested_path,
                        found_path_parts,
                        mapping_type,
                    ), property_path = _look_up_known_parent_paths(
                        f, last_path, property_path, m
                    )
            # else (i.e mapping_type == Text, Keyword, ...) (i.e nested_path with value => in a nested field
            if len(nested_path) >= 1:
                keyword_path = _build_keyword_path(mapping_type, property_path)
                if keyword_path:
                    filter_or_must_or_must_not = (
                        "filter"
                        if f.operator == FilterOperator.EQUAL.value
                        else "must_not"
                    )
                    k_path = keyword_path
                    term_or_match = "terms" if isinstance(f.value, list) else "term"
                else:
                    filter_or_must_or_must_not = (
                        "must"
                        if f.operator == FilterOperator.EQUAL.value
                        else "must_not"
                    )
                    k_path = property_path
                    term_or_match = "match"
                n_path = nested_path

            # nested_path = [] => (i.e mapping_type == Text, Date, Keyword, ...) - nested_path empty
            else:
                found_path = ".".join(found_path_parts) if found_path_parts else None
                if mapping_type is None or (found_path and found_path != property_path):
                    mapping_type = _detect_mapping_type(f.value)
                    keyword_path = _build_keyword_path(
                        None,
                        property_path,
                        dynamic_mapping_type=mapping_type,
                        default_str_keyword_field=default_str_keyword_field
                    )
                else:
                    keyword_path = _build_keyword_path(mapping_type, property_path)
                if keyword_path:
                    filter_or_must_or_must_not = (
                        "filter"
                        if f.operator == FilterOperator.EQUAL.value
                        else "must_not"
                    )
                    k_path = keyword_path
                    term_or_match = "terms" if isinstance(f.value, list) else "term"
                else:
                    filter_or_must_or_must_not = (
                        "must"
                        if f.operator == FilterOperator.EQUAL.value
                        else "must_not"
                    )
                    k_path = property_path
                    term_or_match = "match"
                n_path = None

            _filter, must, must_not, script_score = _build_bool_query(
                f,
                mapping_type,
                k_path,
                property_path,
                filter_or_must_or_must_not,
                term_or_match,
                n_path,
            )

            if _filter:
                es_filters.append(_filter)
            if must:
                musts.append(must)
            if must_not:
                must_nots.append(must_not)
            if script_score:
                script_scores.append(script_score)

        if len(script_scores) > 1:
            raise ValueError(
                f"Multiple dense vector similarity query are not supported: {len(script_scores)} filters involving dense vectors were provided."
            )
        elif len(script_scores) == 1:
            bool_query = _wrap_in_bool_query(es_filters, musts, must_nots)
            return _wrap_in_script_score_query(
                bool_query, script_scores[0], includes, excludes
            )
        else:
            return _wrap_in_bool_query(es_filters, musts, must_nots, includes, excludes)


def _look_up_known_parent_paths(f, last_path, property_path, m):
    if (
        len(f.path) >= 2
        and last_path in ["id", "@id"]
        and f.path[-2] in ["type", "@type"]
    ):  # to cope with paths.type.id TODO: fix forge.paths to not add id after type
        property_path = _join_property_path(f.path[:-2], "@type")

    elif len(f.path) >= 1 and last_path in ["type", "@type"]:
        property_path = _join_property_path(f.path[:-1], "@type")

    elif len(f.path) >= 1 and last_path in ["id", "@id"]:
        property_path = _join_property_path(f.path[:-1], "@id")

    parts = property_path.split(".")
    return _recursive_resolve_nested(m, parts), property_path


def _join_property_path(f_path, to_join_with):
    property_path = ".".join(f_path)
    property_path = (
        property_path + "." + to_join_with if property_path != "" else to_join_with
    )
    return property_path


def _wrap_in_bool_query(filters, musts, must_nots, includes=None, excludes=None):
    bool_query = {
        "query": {"bool": {"filter": filters, "must": musts, "must_not": must_nots}}
    }
    return _add_source(bool_query, includes, excludes)


def _wrap_in_script_score_query(query, script, includes=None, excludes=None):
    # Todo: Catch non supported combination of nested query and non nested ones
    if isinstance(script, elasticsearch_dsl.query.Nested):
        script.query = elasticsearch_dsl.query.ScriptScore(
            query=query["query"], script=script.query.to_dict()["script"]
        )
        script_score_query = script
    else:
        script_score_query = elasticsearch_dsl.query.ScriptScore(
            query=query["query"], script=script.to_dict()["script"]
        )
    query = elasticsearch_dsl.Search().query(script_score_query.to_dict()).to_dict()
    return _add_source(query, includes, excludes)


def _add_source(query, includes, excludes):
    _source = {}
    if includes:
        _source["includes"] = includes
    if excludes:
        _source["excludes"] = excludes
    if _source:
        query_copy = copy.deepcopy(query)
        query_copy["_source"] = _source
        return query_copy
    else:
        return query


def _recursive_resolve_nested(m, field_path):
    nested_path, mapping_type = m.resolve_nested(field_path=".".join(field_path))
    if mapping_type is None and nested_path == ():
        if len(field_path) > 1:
            return _recursive_resolve_nested(
                m, field_path=field_path[0 : len(field_path) - 1]
            )
        else:
            return nested_path, field_path, mapping_type

    else:
        return nested_path, field_path, mapping_type


def _build_keyword_path(
    mapping_type,
    property_path,
    dynamic_mapping_type=None,
    default_str_keyword_field=None,
):
    if isinstance(mapping_type, elasticsearch_dsl.Keyword):
        keyword_path = property_path
    elif (
        mapping_type is None
        and default_str_keyword_field is not None
        and isinstance(dynamic_mapping_type, elasticsearch_dsl.Text)
    ):
        keyword_path = ".".join([property_path, default_str_keyword_field])
    elif mapping_type is not None:
        keyword_fields = [
            keyword_field
            for keyword_field, val in mapping_type.fields.to_dict().items()
            if isinstance(val, elasticsearch_dsl.Keyword)
        ]
        keyword_path = (
            ".".join([property_path, keyword_fields[0]])
            if len(keyword_fields) >= 1
            else None
        )
    else:
        keyword_path = None
    return keyword_path


def _build_bool_query(
    filter: Filter,
    mapping_type: Field,
    k_path: str,
    property_path: str,
    filter_or_must_or_must_not: str,
    term_or_match: str,
    nested_path: List[str] = None,
):
    _filter = None
    must = None
    must_not = None
    script_score = None
    if (
        filter.operator in elasticsearch_operator_range_map.keys()
        and not isinstance(mapping_type, elasticsearch_dsl.Text)
        and not isinstance(mapping_type, elasticsearch_dsl.Boolean)
    ):  # range filter query
        if nested_path:
            _filter = _wrap_in_nested_bool_query(
                nested_path[-1],
                "filter",
                "range",
                {
                    property_path: {
                        elasticsearch_operator_range_map[filter.operator]: filter.value
                    }
                },
            )
        else:
            _filter = _wrap_in_non_nested_query_query(
                property_path,
                "range",
                {elasticsearch_operator_range_map[filter.operator]: filter.value},
            )

    elif filter.operator in elasticsearch_operator_range_map.keys() and (
        isinstance(mapping_type, elasticsearch_dsl.Text)
        or isinstance(mapping_type, elasticsearch_dsl.Boolean)
    ):
        raise ValueError(
            f"Using the range operator {filter.operator} on the Text field/path {filter.path} is not supported."
        )
    elif filter.operator == FilterOperator.EQUAL.value and isinstance(
        mapping_type, elasticsearch_dsl.DenseVector
    ):
        if nested_path:
            sim_query = _wrap_in_script_query(k_path, filter.value)
            script_score = _wrap_in_nested_query(
                path=nested_path[-1], query=sim_query.to_dict()
            )
        else:
            script_score = _wrap_in_script_query(k_path, filter.value)

        must = _wrap_in_non_nested_query_query(
            "field", "exists", k_path
        )  # otherwise script_score will fail. This will be added in a nested script_score query
    else:
        if nested_path:
            query = _wrap_in_nested_bool_query(
                nested_path[-1],
                filter_or_must_or_must_not,
                term_or_match,
                {k_path: filter.value},
            )
        else:
            query = _wrap_in_non_nested_query_query(k_path, term_or_match, filter.value)

        if filter_or_must_or_must_not == "filter":
            _filter = query
        elif filter_or_must_or_must_not == "must":
            must = query
        else:
            must_not = query
    return _filter, must, must_not, script_score


def _wrap_in_nested_bool_query(
    path, filter_or_must_or_must_not, filter_type, filter_value
):
    return _wrap_in_nested_query(
        path=path,
        query=elasticsearch_dsl.query.Q(
            "bool",
            **{
                filter_or_must_or_must_not: elasticsearch_dsl.query.Q(
                    filter_type, **filter_value
                ).to_dict()
            },
        ),
    ).to_dict()


def _wrap_in_nested_query(path: str, query: Dict):
    return elasticsearch_dsl.query.Nested(path=path, query=query)


# TODO: reuse kgforge.core.commons.parser._parse_type
def _detect_mapping_type(value: Any):

    try:
        # integer
        if str(value).isnumeric():
            return elasticsearch_dsl.Long()
        # bool
        if value is True or value is False:
            return elasticsearch_dsl.Boolean()
        # double
        if float(value):
            return elasticsearch_dsl.Float()
    except ValueError as ve:
        pass
    except TypeError as te:
        pass
    try:
        # Date
        if isinstance(dateutil.parser.parse(str(value)), datetime.datetime):
            return elasticsearch_dsl.Date()
    except ParserError as pe:
        return elasticsearch_dsl.Text()


def _wrap_in_non_nested_query_query(path, filter_type, filter_value):
    query = {filter_type: {path: filter_value}}
    return query


def _wrap_in_script_query(field: str, queryVector: List):
    return elasticsearch_dsl.query.Script(
        source=f"doc['{field}'].size() == 0 ? 0 : (cosineSimilarity(params.queryVector, doc['{field}'])+1.0) / 2",
        params={"queryVector": queryVector},
    )
